#ifndef lint
#ifdef sccs
static char     sccsid[] = "@(#)XSunIMSj3.c	1.01 92/06/02 SMI";
#endif
#endif

/*
****************************************************************************

              Copyright 1990,1991, by Sun Microsystems, Inc.
              Copyright 1991, by FUJITSU LIMITED

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of FUJITSU LIMITED and Sun
Microsystems, Inc. not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.  Sun Microsystems, Inc. and FUJITSU LIMITED make no
representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.

SUN MICROSYSTEMS, INC. AND FUJITSU LIMITED DISCLAIM ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL SUN MICROSYSTEMS, INC.
AND FUJITSU LIMITED BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

Auther: Hideki Hiura	Sun Microsystems, Inc.
        Riki Kawaguchi	FUJITSU LIMITED


****************************************************************************
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#ifdef SVR4
#include <sys/utsname.h>
#include <sys/systeminfo.h>
#endif /* SVR4 */
#include <locale.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#if XlibSpecificationRelease != 5
#include <X11/XlibR5.h>
#endif /* XlibSpecificationRelease */
#include "XSunExt.h"
#include <X11/keysym.h>
#include <X11/Xutil.h>

#include "XSunIMProt.h"
#include "XSunIMPriv.h"
#include "XSunIMMMan.h"
#include "XSunIMMthd.h"
#include "XSunIMPub.h"
#include "XSunIMTrans.h"
#include "XSunIMSj3.h"
#include "sj3lib.h"

Private Bool	iml_sj3_OpenIF	    (/* iml_if_t *If */);
Private Bool 	iml_sj3_CloseIF	    (/* iml_if_t *If */);
Private Bool 	iml_sj3_GetIFValue  (/* iml_if_t *If, char *atr, void *val */);
Private Bool 	iml_sj3_SetIFValue  (/* iml_if_t *If, char *atr, void *val */);
Private Bool	iml_sj3_DestroySC   (/* iml_session_t *s */);
Private void	iml_sj3_SetSCFocus  (/* iml_session_t *s */);
Private void	iml_sj3_UnsetSCFocus(/* iml_session_t *s */);
Private void	iml_sj3_SendEvent   (/* iml_session_t *s, XEvent * */);
Private char *	iml_sj3_ResetSC	    (/* iml_session_t *s */);
Private Bool	iml_sj3_GetSCValue  (/* iml_session_t *s, char *atr, void *val */);
Private Bool	iml_sj3_SetSCValue  (/* iml_session_t *s, char *atr, void *val */);
Private iml_session_t * iml_sj3_CreateSC    (/* iml_if_t *If, void *ic, char *le_name */);

#ifdef DYNAMICIF
Private
#else
Public
#endif
if_methods_t sj3_methods = {
    iml_sj3_OpenIF,
    iml_sj3_CloseIF,
    iml_sj3_GetIFValue,
    iml_sj3_SetIFValue,
    (void *(*)())iml_sj3_CreateSC,
    iml_sj3_DestroySC,
    iml_sj3_SetSCValue,
    iml_sj3_GetSCValue,
    iml_sj3_ResetSC,
    iml_sj3_SetSCFocus,
    iml_sj3_UnsetSCFocus,
    iml_sj3_SendEvent
};

Private void	trans_sj3_Commit();
Private void	trans_sj3_Backspace();
Private void	trans_sj3_Conversion();
Private void	trans_sj3_Return();
Private void	trans_sj3_Conv();
Private void	trans_sj3_Nextword();
Private void	trans_sj3_Backword();
Private void	trans_sj3_RomaConv();
Private void	trans_sj3_Unmatch();
Private void	trans_sj3_ZAlfaNum();
Private void	trans_sj3_HiraganaMode();
Private void	trans_sj3_KatakanaMode();
Private void	trans_sj3_LookupChoice();

#define NumOfTransMethod	(sizeof(transmethod)/sizeof(transmethod[0]))

Private void	 sj3x_opsw_init		 (/* */);
Private void	 sj3x_send_commit	 (/* iml_session_t *s */);
Private void	 sj3x_eval_results	 (/* iml_session_t *s, iml_inst *lp */);
Private void	 sj3x_make_conversion_off(/* iml_session_t *s */);
Private void	 sj3x_make_conversion_on (/* iml_session_t *s */);
Private void	 sj3x_keypress		 (/* iml_session_t *s */);
Private void	 sj3x_rkconvs		 (/* iml_session_t *s */);
Private void	 sj3x_set_choice	 (/* iml_session_t *s, int selected */);
Private void	 init_status_string	 (/* iml_session_t *s */);

Private void	 sj3_AddPreeditStrings();
Private void	 sj3_AddSJISStrings();
Private void	 sj3_CopyFeedbacks(/* sj3_feedback *, XIMFeedback, int */);
Private void	 sj3_SetBunsetu(/* sj3_bunsetu *, struct bunsetu *, int */);
Private void	 sj3_AllocDouonBuff(/* sj3_douon *, int */);
Private void	 sj3_SetFeedback(/* iml_session_t *, sj3_feedback *, sj3_bunsetu * */);
Private void	 SetPreedit(/* iml_session_t *, sj3_bunsetu *, char *, int */);
Private void	 ReSetBunsetu(/* sj3_SJISstring *, struct bunsetu *, int */);
Private void	 GetDouon(/* struct bunsetu * */);
Private int	 ShiftJisToEuc(/* uchar *sjis, int sjlen, wchar_t *euc, int len */);
Private String	 HiraganaToKatakana();
Private wchar_t	 	sjis_to_jis(/* unsigned short code */);
Private unsigned short  jis_to_sjis(/* unsigned short */);

/* Status Area */
Private wchar_t		StatusString[SJ3Status_MAX_NUM][SJ3Status_MAX_COLUMS];

#ifdef DEBUG
Private void
_memd( ptr, size )
    unsigned char *ptr;
    int	  size;
{
    int i;
    for (i=0;i<size;i++,ptr++) {
	printf( " %2.2x", *ptr );
    }
    printf( "\n" );
}
#endif

Private Bool
iml_sj3_OpenIF( If )
    iml_if_t	*If;
{
    sj3x_opsw_init();
    return True;
}

Private Bool
iml_sj3_CloseIF( If )
    iml_if_t	*If;
{
    return True;
}

Private Bool
iml_sj3_GetIFValue( If, attr, value )
    iml_if_t	*If;
    char	*attr;
    void	*value;
{
    if ( !strcmp( attr, IF_VERSION ) ) {
	*(char **)value = SJ3_IF_VERSION;
	return True;
    } else if ( !strcmp( attr, IF_LE_NAME ) ) {
	*(char **)value = SJ3_IF_LE_NAME;
	return True;
    }
    return False;
}

Private Bool
iml_sj3_SetIFValue( If, attr, value )
    iml_if_t	*If;
    char	*attr;
    void	*value;
{
    return False;
}

Private iml_session_t *
iml_sj3_CreateSC( If, ic, le_name )
    iml_if_t	*If;
    void	*ic;
    char	*le_name;
{
    Private int	 sj3x_create_session(
#if NeedFunctionPrototypes
	int		 id,
	iml_session_t	*s,
	Bool		 is_recycled
#endif		/* NeedFunctionPrototypes */
    );
    int		 le_id;

    le_id = 0;
    return If->m->iml_construct_session( If, ic, le_id, sizeof(iml_session_t), sj3x_create_session );
}

Private Bool
iml_sj3_DestroySC( s )
    iml_session_t	*s;
{

}

Private char *
iml_sj3_ResetSC( s )
    iml_session_t	*s;
{
    sj3x_send_commit(s);
}

Private Bool
iml_sj3_SetSCValue( s, attr, value )
    iml_session_t	*s;
    char		*attr;
    void		*value;
{
    if ( !strcmp( attr, SC_Henkan_Mode ) ||
	 !strcmp( attr, XNExtXimp_Conversion ) ) {
	if ( (Bool)value == True ) {
	    if ( !( s->public_status & IMLSTATUS_Henkan_Mode ) ) {
		s->public_status |= IMLSTATUS_Henkan_Mode;
		sj3x_make_conversion_on( s );
	    }
	} else {
	    if ( s->public_status & IMLSTATUS_Henkan_Mode ) {
		s->public_status &= ~IMLSTATUS_Henkan_Mode;
		sj3x_make_conversion_off( s );
	    }
	}
	return True;
    } else if ( !strcmp( attr, SC_STATUS ) ) {
	if ( !( s->public_status & IMLSTATUS_Henkan_Mode ) ) {
	    if ( (iml_status_t)value & IMLSTATUS_Henkan_Mode ) {
		sj3x_make_conversion_on( s );
	    }
	}
	s->public_status = (iml_status_t)value;
	return True;
    }
    return False;
}

Private Bool
iml_sj3_GetSCValue( s, attr, value )
    iml_session_t	*s;
    char		*attr;
    void		*value;
{
    if ( !strcmp( attr, SC_STATUS ) ) {
	iml_inst	*lp = s->If->m->iml_make_status_notify_inst( s );
	*(iml_status_t *)value = s->public_status;
	if ( lp = s->If->m->iml_execute( s, &lp ) ) {
	    sj3x_eval_results( s, lp );
	}
	return True;
    }
    return False;
}

Private void
iml_sj3_SetSCFocus( s )
    iml_session_t	*s;
{
    wchar_t		*status_string;
    iml_inst	*lp;
    iml_inst	*rrv = NULL;

    lp = s->If->m->iml_make_status_notify_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    status_string = GetStatus( s );
    lp = s->If->m->iml_make_status_draw_inst( s, status_string, NULL );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

Private void
iml_sj3_UnsetSCFocus( s )
    iml_session_t	*s;
{
    iml_inst	*lp = s->If->m->iml_make_status_notify_inst( s );

    if ( lp = s->If->m->iml_execute( s, &lp ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

#define sj3x_is_conv_on_key(s,len) ((len)>0 && s->keysym == XK_space && (s)->event->state & ControlMask )

Private void
iml_sj3_SendEvent( s, e )
    iml_session_t	*s;
    XKeyEvent		*e;
{
    iml_inst	*lp;
    int		 len;

    if ( e && e->type != KeyPress ) {
	return;
    }
    if ( e ) {
	s->event = e;
    } else {
	e = s->event;
    }
    s->XLookupBuf[0] = '\0';
    len = s->If->m->iml_lookupString( e,
			s->XLookupBuf, s->XLookupBuflen, &s->keysym );
    if ( !( s->public_status & IMLSTATUS_Henkan_Mode ) ) {
	/*
	 * Conversion off
	 */
	if ( ( s->keysym == XK_Henkan_Mode ) ||
	       sj3x_is_conv_on_key( s, len ) ) {
	    sj3x_make_conversion_on( s );
	    return;
	}
	lp = s->If->m->iml_make_keypress_inst( s );
	if ( lp = s->If->m->iml_execute( s, &lp ) ) {
	    sj3x_eval_results( s, lp );
	}
	return;
    }
    /*
     * Conversion on
     */
    if ( sj3x_is_key( s ) ) {
	sj3x_keypress( s, len );
    }
    return;
}

Private void
sj3x_eval_results( s, results )
    iml_session_t	*s;
    iml_inst		*results;
{
    XIMLookupStartCallbackStruct	*ls;
    XIMLookupProcessCallbackStruct	*lp;

    if ( results->opcode & IMM_CB_RESULT_REQUIRED ) {
	results->opcode &= ~IMM_CB_RESULT_REQUIRED;
	switch ( results->opcode ) {
	case IMM_LOOKUP_START:
	    ls = (XIMLookupStartCallbackStruct *)& results->operand;
	    if ( s->WhoIsMaster == HasNotBeenNegotiated ) {
		s->WhoIsMaster = ls->WhoIsMaster;
	    } else {
		break;
	    }
	    s->XIMPreference = s->CBPreference;
	    break;
	case IMM_LOOKUP_PROCESS:
	    lp = (XIMLookupProcessCallbackStruct *)& results->operand;
	    if ( ( lp->index_of_choice_selected == XIM_UNDETERMINED ) ||
		 ( lp->index_of_choice_selected == XIM_UNKNOWN_KEYSYM ) ) {
		break;
	    }
	    sj3x_set_choice( s, lp->index_of_choice_selected );
	    break;
	case IMM_PREEDIT_START:
	    s->PreEditBufferSizeInCallback = *(int *)& results->operand;
	    break;
	default:
	    break;
	}
    }
    return;
}

Private void
sj3x_make_conversion_off( s )
    iml_session_t	*s;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    wchar_t		*status_string;

    s->convMode = SJ3Status_CONVOFF;

    status_string = GetStatus( s );
    lp = s->If->m->iml_make_status_draw_inst( s, status_string, NULL );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    lp = s->If->m->iml_make_preedit_done_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    s->public_status &= ~IMLSTATUS_Henkan_Mode;
    lp = s->If->m->iml_make_status_notify_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }

    return;
}

Private void
sj3x_make_conversion_on( s )
    iml_session_t	*s;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    wchar_t		*status_string;

    s->public_status |= IMLSTATUS_Henkan_Mode;
    s->convMode = SJ3Status_HIRAGANA;

    lp = s->If->m->iml_make_status_notify_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    status_string = GetStatus( s );
    lp = s->If->m->iml_make_status_draw_inst( s, status_string, NULL );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    lp = s->If->m->iml_make_preedit_start_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    s->num_of_conved = 0;
    s->current_bnum = 0;
    return;
}

#define	SJ3_UNAME		"Htt"
#define SJ3_ENV			"SJ3ENV"
#define SJ3_BIND_NAME		"htt.sj3.bind"
#define SJ3_BIND_CLASS		"Htt.Sj3.Bind"

Private HttActionRec		actionRec[] = {
    {(String )"conversion_on", trans_sj3_Conversion },
    {(String )"conversion_off", trans_sj3_Conversion },
    {(String )"commit", trans_sj3_Commit },
    {(String )"backspace", trans_sj3_Backspace },
    {(String )"nextword", trans_sj3_Nextword },
    {(String )"backword", trans_sj3_Backword },
    {(String )"conversion", trans_sj3_Conv },
    {(String )"return", trans_sj3_Return },
    {(String )"romaconv", trans_sj3_RomaConv },
    {(String )"alfanumMode", trans_sj3_ZAlfaNum },
    {(String )"hiraganaMode", trans_sj3_HiraganaMode },
    {(String )"katakanaMode", trans_sj3_KatakanaMode },
    {(String )"lookupChoice", trans_sj3_LookupChoice },
};

Private int
sj3x_create_session( le_id, s, is_recycled )
    int			 le_id;
    iml_session_t	*s;
    Bool		 is_recycled;
{
    int			 retcd;
    char		*host_name = NULL;
    char		*user_name = SJ3_UNAME;
    HttTCid		 tcid;
    static int		 session_id = 0;

    if ( session_id == 0 ) {
	if ( ( retcd = sj3_open( host_name, user_name ) ) < 0 ) {
	    fprintf( stderr, "sj3 I/F: language engine (id %d) open failed.\n", le_id );
	    return -1;
	}
	init_status_string( s );
    }

    /*
     * create Translation Context
     */
    tcid = Htt_CreateTC( s );
    if ( !tcid ) {
	fprintf( stderr, "sj3 I/F: translation (id %d) create failed.\n", le_id );
	return -1;
    }

    s->tc = tcid;

    if ( !is_recycled ) {
	s->romaString.area_size = 0 ;
	s->convString.area_size = 0 ;
	s->bunsetu.count = 0;
	s->douon.count = 0;
	s->preeditString.count = 0 ;
	s->preeditFeedback.alloc_count = 0 ;

	s->romaString.string = (String )NULL;
	s->convString.string = (String )NULL;
	s->bunsetu.bun = (struct bunsetu *)NULL;
	s->douon.douon = (struct douon *)NULL;
	s->preeditString.string = (wchar_t *)NULL;
	s->preeditFeedback.feedback = (XIMFeedback *)NULL;
    }

    s->romaString.use_byte = 0;
    s->convString.use_byte = 0;
    s->bunsetu.use_count = 0;
    s->douon.use_count = 0;
    s->preeditString.use_count = 0;
    s->preeditFeedback.use_count = 0;

    s->tc->tm->tm_make_translations( s->tc, SJ3_BIND_NAME, SJ3_BIND_CLASS );
    s->tc->tm->tm_add_action( s->tc, actionRec, XtNumber( actionRec ) );
    s->tc->tm->tm_unmatch_action( s->tc, trans_sj3_Unmatch, NULL, 0 );

    if ( ++session_id < 0 ) {
	session_id = 1;
    }
    return session_id;
}

Private void
init_status_string( s )
    iml_session_t	*s;
{
    /*
     * FIX ME !!
     */
    static unsigned short	_convoff[] = SJ3String_CONVOFF;
    static unsigned short	_hiragana[] = SJ3String_HIRAGANA;
    static unsigned short	_katakana[] = SJ3String_KATAKANA;
    static unsigned short	_alfanum[] = SJ3String_ALFANUM;

    mbstowcs( (wchar_t *)StatusString[SJ3Status_CONVOFF],
			(char *)_convoff, strlen(_convoff) );
    mbstowcs( (wchar_t *)StatusString[SJ3Status_HIRAGANA],
			(char *)_hiragana, strlen(_hiragana) );
    mbstowcs( (wchar_t *)StatusString[SJ3Status_KATAKANA],
			(char *)_katakana, strlen(_katakana) );
    mbstowcs( (wchar_t *)StatusString[SJ3Status_ALFANUM],
			(char *)_alfanum, strlen(_alfanum) );
    return;
}

Private void
sj3x_ZAlfaNum( s, e )
    iml_session_t	*s;
    XKeyEvent		*e;
{
    unsigned short	 args[2];
    String		 zenkaku;
    wchar_t		*romaBuff;

    args[0] = GetZenkaku( s->XLookupBuf[0] );
    args[1] = '\0';
    zenkaku = (String )args;
    trans_sj3_RomaConv( s, e, &zenkaku, 1 );
    return;
}

Private Bool
sj3x_is_key( s )
    iml_session_t	*s;
{
    switch ( s->keysym ) {
    case XK_Control_L:
    case XK_Control_R:
    case XK_Meta_L:
    case XK_Meta_R:
    case XK_Alt_R:
    case XK_Alt_L:
    case XK_Super_R:
    case XK_Super_L:
    case XK_Hyper_R:
    case XK_Hyper_L:
	return False;
    }
    return True;
}

Private int
Sj3StateToTMState( s )
    iml_session_t	*s;
{
    if ( Is_lookupMode( s ) )
	return TM_ParseState_LookupChoice;
    if ( Is_conved( s ) )
	return TM_ParseState_PreeditConv;
    return TM_ParseState_PreeditEdit;
}

Private void
sj3x_keypress( s )
    iml_session_t	*s;
{
    iml_inst	*lp;

    if ( Is_lookupMode( s ) ) { /* Lookup Choice */
	lp = s->If->m->iml_make_lookup_process_inst( s, True );
	if ( lp = s->If->m->iml_execute( s, &lp ) ) {
	    sj3x_eval_results( s, lp );
	}
	return;
    }
    if ( isprint( s->XLookupBuf[0] ) ) {
	if ( s->num_of_conved != 0 || s->current_bnum != 0 ) {
	    trans_sj3_Commit( s, (XEvent *)NULL, (String *)NULL, 0 );
	}
	switch ( s->convMode ) {
	case SJ3Status_HIRAGANA:
	case SJ3Status_KATAKANA:
	    s->tc->tm->tm_add_eventSeq( s->tc, s->event );
	    (void )s->tc->tm->tm_match_event( s->tc, Sj3StateToTMState( s ) );
	    break;
	case SJ3Status_ALFANUM:
	    sj3x_ZAlfaNum( s, (XKeyEvent *)s->event );
	    break;
	}
    } else {
	s->tc->tm->tm_add_eventSeq( s->tc, s->event );
	(void )s->tc->tm->tm_match_event( s->tc, Sj3StateToTMState( s ) );
    }
    return;
}

Private void
CommitSimpleCode( s, c )
    iml_session_t	*s;
    char		 c;
{
    iml_inst		*lp;
    iml_inst		*rrv = NULL;

    lp = s->If->m->iml_make_commit_inst( s, &c, 1 );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

Private void
sj3x_set_choice( s, selected )
    iml_session_t	*s;
    int			 selected;
{
    iml_inst	*lp;
    iml_inst	*rrv = NULL;
    wchar_t	*tmp;
    int		 wcssize;
    int		 word_count;

    if ( selected < 0 || selected >= NumOfDouon(s) )
	return;
    s->is_lookup = False;
    lp = s-> If->m->iml_make_lookup_done_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    SetPreedit( s, &s->bunsetu,
		s->douon.douon[selected].ddata, s->douon.douon[selected].dlen );
    DisplayPreedit( s, rrv, lp );
    return;
}

Private void
trans_sj3_Conversion( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;

    sj3x_make_conversion_off( s );
    return;
}

Private void
trans_sj3_RomaConv( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    int			 size;
    int			 word_count;
    int			 sjis_count;
    int			 i;
    int			 length;
    String		 roma;
    wchar_t		*roma_buff;
    unsigned char	*sjis_buff;
    iml_inst		*rrv = NULL;
    iml_inst		*lp;

    size = strlen( args[0] );
    if ( size == 0 ) {
	return;
    }
    if ( s->convMode == SJ3Status_KATAKANA ) {
	roma = HiraganaToKatakana( args[0], size );
    } else {
	roma = (String )calloc( size + sizeof(char), sizeof(char) );
	strcpy( roma, args[0] );
    }
    roma_buff = (wchar_t *)calloc( ( size + 1 ),  sizeof(wchar_t) );
    mbstowcs( roma_buff, (char *)roma, size );
    word_count = wslen( roma_buff );

    sjis_buff = (unsigned char *)calloc(( size + 1 ), sizeof(char) );
    bzero( sjis_buff, size + 1 );
    EucToShiftJis( roma_buff, word_count, sjis_buff, size );

    sj3_AddSJISStrings( &s->romaString, sjis_buff, size );
    sj3_AddPreeditStrings( &s->preeditString, roma_buff, word_count );
    sj3_CopyFeedbacks( &s->preeditFeedback, XIMReverse, word_count );
    lp = s->If->m->iml_make_preedit_draw_inst( s, s->preeditString.string,
				s->preeditFeedback.feedback );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    free( roma_buff );
    free( roma );
    return;
}

Private void
trans_sj3_Unmatch( s, e, args, num_args )
    iml_session_t	*s;
    XKeyEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*lp;
    iml_inst		*rrv = NULL;
    int			 count;
    char		 preString[10];
    char		 comming[10];
    KeySym		 preKeysym;
    KeySym		 comKeysym;
    String		 zenkaku;
    unsigned short	 word1[2] ;
    unsigned short	 word2[2] ;
    unsigned short	 word3[2] ;

    word1[0] = SmallTSU;
    word1[1] = 0;
    word2[0] = NN;
    word2[1] = 0;
    word3[0] = 0;
    word3[1] = 0;

    preKeysym = comKeysym = 0;
    if ( s->tc->tm->tm_event_to_String( s->tc,
			preString, sizeof(preString), &preKeysym ) ) {
	if ( iscntrl( preString[0] ) )
	    return;
	bzero( comming, sizeof(comming) );
	count = s->If->m->iml_lookupString( e,
			comming, sizeof(comming), &comKeysym );
	if ( preKeysym == comKeysym ) {
	    String	smallTSU = (String )word1;
	    trans_sj3_RomaConv( s, e, &smallTSU, 1 );
	} else if ( preString[0] == 'n' || preString[0] == 'N' ) {
	    String	nN = (String )word2;
	    trans_sj3_RomaConv( s, e, &nN, 1 );
	} else {
	    word3[0] = GetZenkaku( preString[0] );
	    zenkaku = (String )word3;
	    trans_sj3_RomaConv( s, e, &zenkaku, 1 );
	}
    } else {
	bzero( comming, sizeof(comming) );
	s->If->m->iml_lookupString( e, comming, sizeof(comming), &comKeysym );
	if ( iscntrl( comming[0] ) ) {
	    if ( s->preeditString.use_count == 0 ) {
		lp = s->If->m->iml_make_commit_inst( s, 1, &comming[0] );
		s->If->m->iml_link_inst_tail( &rrv, lp );
		if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
		    sj3x_eval_results( s, lp );
		}
	    }
	} else {
	    word3[0] = GetZenkaku( comming[0] );
	    zenkaku = (String )word3;
	    trans_sj3_RomaConv( s, e, &zenkaku, 1 );
	}
    }
    return;
}

Private void
trans_sj3_Conv( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    String		 sjis_buff;
    wchar_t		*wc_buff;
    struct bunsetu	 bun[MAX_BUNSETU];
    char		*p;
    int			 bnum;
    int			 size;
    int			 wc_size;
    int			 word_count;

    if ( !Is_foundConversionStrings(s) ) {
	return;
    }
    if ( Is_notConved(s) ) { /* first conved */
	/*
	 * Ikkatsu Henkan
	 */
	size = 1024;
	sjis_buff = (String )calloc( size, sizeof(char));
	bzero( sjis_buff, size );
	if ( ( s->bnum = sj3_getkan( s->romaString.string, bun,
					sjis_buff, size ) ) == -1 ) {
	    fprintf( stderr, "sj3_getkan: server not connect.\n" );
	    return;
	}
	bnum = s->bnum - 1;
	if ( bun[bnum].destlen == 0 ) {
	    strcat( sjis_buff, bun[bnum].srcstr );
	} else {
	    p = (char *)bun[bnum].deststr;
	    p += bun[bnum].destlen;
	    *p = '\0';
	}

	s->convString.use_byte = 0;
	sj3_AddSJISStrings( &s->convString, sjis_buff, size );

	ReSetBunsetu( &s->convString, bun, s->bnum );
	sj3_SetBunsetu( &s->bunsetu, bun, s->bnum );

	wc_size = strlen( sjis_buff ) * sizeof(wchar_t);
	wc_buff = (wchar_t *)calloc(wc_size, sizeof(char) );
	word_count = ShiftJisToEuc( sjis_buff,
			strlen( sjis_buff ), wc_buff, wc_size );
	s->preeditString.use_count = 0;
	s->current_bnum = 0;
	sj3_AddPreeditStrings( &s->preeditString, wc_buff, word_count );
	sj3_SetFeedback( s, &s->preeditFeedback, &s->bunsetu );
	free( wc_buff );
	free( sjis_buff );

	/*
	 * get douon!!
	 */
	GetDouon( &s->douon, &bun[0] );
	s->num_of_conved = 2;
    } else { /* */
	if ( s->num_of_conved > s->douon.use_count ) {
	    p = (String )s->bunsetu.bun[s->current_bnum].srcstr;
	    size = s->bunsetu.bun[s->current_bnum].srclen;
	    s->num_of_conved = 1;
	} else {
	    bnum = s->num_of_conved - 1;
	    p = (String )s->douon.douon[bnum].ddata;
	    size = s->douon.douon[bnum].dlen;
	    s->num_of_conved++;
	}
	SetPreedit( s, &s->bunsetu, p, size );
    }
    lp = s->If->m->iml_make_preedit_draw_inst( s,
		    s->preeditString.string, s->preeditFeedback.feedback );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	    sj3x_eval_results( s, lp );
    }
    return;
}

Private void
trans_sj3_Nextword( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;

    if ( s->current_bnum >= ( s->bnum - 1 ) )
	return;

    s->current_bnum++;
    GetDouon( &s->douon, &s->bunsetu.bun[s->current_bnum] );
    sj3_SetFeedback( s, &s->preeditFeedback, &s->bunsetu );
    s->num_of_conved = 2;

    lp = s->If->m->iml_make_preedit_draw_inst( s,
		    s->preeditString.string, s->preeditFeedback.feedback );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	    sj3x_eval_results( s, lp );
    }
    return;
}

Private void
trans_sj3_Backword( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;

    if ( s->current_bnum <= 0 )
	return;

    s->current_bnum--;
    GetDouon( &s->douon, &s->bunsetu.bun[s->current_bnum] );
    sj3_SetFeedback( s, &s->preeditFeedback, &s->bunsetu );
    s->num_of_conved = 1;

    lp = s->If->m->iml_make_preedit_draw_inst( s,
		    s->preeditString.string, s->preeditFeedback.feedback );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	    sj3x_eval_results( s, lp );
    }
    return;
}

Private void
trans_sj3_Commit( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    unsigned char	*mbsbuf;
    int			 len;
    int			 size = sizeof(wchar_t) * s->preeditString.use_count;
    int			 alloc_size = size + sizeof(char);

    if ( s->preeditString.use_count == 0 ) {
	return;
    }
    mbsbuf = (unsigned char *)calloc( alloc_size, sizeof(char) );
    len = wcstombs( (char *)mbsbuf,
	    (wchar_t *)s->preeditString.string, size );
    len = strlen( mbsbuf );
    lp = s->If->m->iml_make_preedit_erase_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    lp = s->If->m->iml_make_commit_inst( s, len, mbsbuf );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    s->bnum = 0;
    s->current_bnum = 0;
    s->romaString.use_byte = 0;
    s->convString.area_size = 0;
    s->bunsetu.use_count = 0;
    s->douon.use_count = 0;
    s->preeditString.use_count = 0;
    s->preeditFeedback.use_count = 0;
    s->num_of_conved = 0;
    free( mbsbuf );
    return;
}

Private void
trans_sj3_Return( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;

    trans_sj3_Commit( s, e, args, num_args );
    lp = s->If->m->iml_make_commit_inst( s, 1, &s->XLookupBuf[0] );
    s->If->m->iml_link_inst_tail( &rrv, lp );
    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

/*
 * FIX ME !! can do 2byte-code only...
 */
Private void
trans_sj3_Backspace( s, e, args, num_args )
    iml_session_t	*s;
    XKeyEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*lp;
    iml_inst		*rrv = NULL;

    if ( Is_conved( s ) ) {
	trans_sj3_Commit( s, e, args, num_args );
    }
    if ( Is_foundConversionStrings( s ) ) {
	DelSJISString( &s->romaString );
	DelPreeditString( &s->preeditString );
	DelFeedback( &s->preeditFeedback );
	DisplayPreedit( s, rrv, lp );
    } else {
	CommitSimpleCode( s, s->XLookupBuf[0] );
    }
    return;
}

Private void
trans_sj3_LookupChoice( s, e, args, num_args )
    iml_session_t	*s;
    XKeyEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*lp;
    String		*strList;
    int			 i;
    register struct douon	*douon;

    if ( Is_notConved(s) )
	trans_sj3_Conv( s, e, NULL, 0 );
    if ( NumOfDouon( s ) == 0 )
	return;
    douon = s->douon.douon;
    s->is_lookup = True;
    lp = s->If->m->iml_make_lookup_start_inst( s );
    if ( lp = s->If->m->iml_execute( s, &lp ) ) {
	sj3x_eval_results( s, lp );
    }
    strList = (String *)calloc( NumOfDouon(s), sizeof(String *) );
    for ( i = 0; i < NumOfDouon(s); i++, douon++ ) {
	int	     wcssize = ( douon->dlen + 1 ) * sizeof(wchar_t);
	int	     mbssize = ( douon->dlen + 1 ) * MB_LEN_MAX;
	int	     word_count;
	wchar_t	    *tmp = (wchar_t *)calloc(douon->dlen + 1, sizeof(wchar_t) );
	strList[i] = (String )calloc( mbssize, sizeof(char) );
	word_count = ShiftJisToEuc( douon->ddata, douon->dlen, tmp, wcssize );
	wcstombs( strList[i], tmp, word_count * sizeof(wchar_t) );
	free( tmp );
    }
    lp = s->If->m->iml_make_lookup_draw_inst( s, strList, NumOfDouon(s) );
    if ( lp = s->If->m->iml_execute( s, &lp ) ) {
	sj3x_eval_results( s, lp );
    }
    for ( i = 0; i < NumOfDouon(s); i++ )
	free( strList[i] );
    free( strList );
    return;
}

Private void
trans_sj3_HiraganaMode( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    wchar_t		*statusString;

    if ( s->convMode == SJ3Status_HIRAGANA )
	return;
    if ( Is_foundConversionStrings(s) ) { /* Already convert */
	trans_sj3_Commit( s, e, args, num_args );
    }
    s->convMode = SJ3Status_HIRAGANA;

    lp = s->If->m->iml_make_status_notify_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    statusString = GetStatus( s );
    lp = s->If->m->iml_make_status_draw_inst( s, statusString, NULL );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    lp = s->If->m->iml_make_preedit_start_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

Private void
trans_sj3_KatakanaMode( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    wchar_t		*statusString;

    if ( s->convMode == SJ3Status_KATAKANA )
	return;
    if ( Is_foundConversionStrings(s) ) { /* Already convert */
	trans_sj3_Commit( s, e, args, num_args );
    }
    s->convMode = SJ3Status_KATAKANA;

    lp = s->If->m->iml_make_status_notify_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    statusString = GetStatus( s );
    lp = s->If->m->iml_make_status_draw_inst( s, statusString, NULL );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    lp = s->If->m->iml_make_preedit_start_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

Private void
trans_sj3_ZAlfaNum( s, e, args, num_args )
    iml_session_t	*s;
    XEvent		*e;
    String		*args;
    int			 num_args;
{
    iml_inst		*rrv = NULL;
    iml_inst		*lp;
    wchar_t		*statusString;

    if ( s->convMode == SJ3Status_ALFANUM )
	return;
    if ( Is_foundConversionStrings(s) ) { /* Already convert */
	trans_sj3_Commit( s, e, args, num_args );
    }
    s->convMode = SJ3Status_ALFANUM;

    lp = s->If->m->iml_make_status_notify_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    statusString = GetStatus( s );
    lp = s->If->m->iml_make_status_draw_inst( s, statusString, NULL );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    lp = s->If->m->iml_make_preedit_start_inst( s );
    s->If->m->iml_link_inst_tail( &rrv, lp );

    if ( lp = s->If->m->iml_execute( s, &rrv ) ) {
	sj3x_eval_results( s, lp );
    }
    return;
}

Private void
sj3x_send_commit( s )
    iml_session_t *s;
{

}

Private void
sj3x_opsw_init()
{

}

Private void
sj3_AddSJISStrings( sjisP, sjis, size )
    sj3_SJISstring	*sjisP;
    unsigned char	*sjis;
    int			 size;
{
    unsigned char	*old = (unsigned char *)sjisP->string;
    int			 needed_size = sjisP->use_byte + size + 1;
    if ( size == 0 )
	return;
    if ( needed_size > sjisP->area_size ) {
	int	alloc_count = needed_size / SJ3_STRING_ALLOC_SIZE;
	alloc_count += ((needed_size % SJ3_STRING_ALLOC_SIZE) ? 1 : 0);
	sjisP->area_size = SJ3_STRING_ALLOC_SIZE * alloc_count;
	sjisP->string = (char *)calloc( sjisP->area_size, sizeof(char) );
	if ( old ) {
	    bcopy( (char *)old, (char *)sjisP->string, sjisP->use_byte );
	    free( old );
	}
    }
    bcopy( (char *)sjis, (char *)(sjisP->string + sjisP->use_byte), size );
    sjisP->use_byte += size;
    sjisP->string[sjisP->use_byte] = '\0';
    return;
}

Private void
sj3_AddPreeditStrings( eucp, euc, size )
    sj3_EUCstring	*eucp;
    wchar_t		*euc;
    int			 size;
{
    wchar_t		*old = eucp->string;
    int			 needed_size = eucp->use_count + size + 1;
    if ( size == 0 )
	return;
    if ( needed_size > eucp->count ) {
	int	alloc_count = needed_size / SJ3_STRING_ALLOC_SIZE;
	alloc_count += ((needed_size % SJ3_STRING_ALLOC_SIZE) ? 1 : 0);
	eucp->count = SJ3_STRING_ALLOC_SIZE * alloc_count;
	eucp->string = (wchar_t *)calloc( eucp->count, sizeof(wchar_t) );
	if ( old ) {
	    bcopy( (char *)old, (char *)eucp->string,
			eucp->use_count * sizeof(wchar_t) );
	    free( old );
	}
    }
    bcopy( (char *)euc,
	(char *)(eucp->string + eucp->use_count), size * sizeof(wchar_t) );
    eucp->use_count += size;
    eucp->string[eucp->use_count] = (wchar_t)0;
    return;
}

Private void
sj3_CopyFeedbacks( fbp, feedback, copies )
    sj3_feedback	*fbp;
    XIMFeedback		 feedback;
    int			 copies;
{
    XIMFeedback		*old = fbp->feedback;
    XIMFeedback		*feedbackP;
    int			 needed_size = fbp->use_count + copies;
    int			 i;

    if ( copies == 0 )
	return;
    if ( needed_size > fbp->alloc_count ) {
	int	alloc_count = needed_size / SJ3_STRING_ALLOC_SIZE;
	alloc_count += ((needed_size % SJ3_STRING_ALLOC_SIZE) ? 1 : 0);
	fbp->alloc_count = SJ3_STRING_ALLOC_SIZE * alloc_count;
	fbp->feedback = (XIMFeedback * )
			    calloc( fbp->alloc_count, sizeof(XIMFeedback) );
	if ( old ) {
	    bcopy( (char *)old, (char *)fbp->feedback,
			fbp->use_count * sizeof(XIMFeedback) );
	    free( old );
	}
    }
    feedbackP = fbp->feedback + fbp->use_count;
    for ( i = 0; i < copies; i++, feedbackP++ )
	*feedbackP = feedback;
    fbp->use_count += copies;
    return;
}


Private void
sj3_AllocDouonBuff( douonP, count )
    sj3_douon		*douonP;
    int			 count;
{
    struct douon	*old = douonP->douon;
    int			 needed_size = count;

#define SJ3_DOUON_ALLOC_SIZE	8
    if ( count == 0 )
	return;
    if ( ( needed_size ) > douonP->count ) {
	int	alloc_size = needed_size / SJ3_DOUON_ALLOC_SIZE;
	alloc_size += ((needed_size % SJ3_DOUON_ALLOC_SIZE) ? 1 : 0);
	douonP->count = SJ3_DOUON_ALLOC_SIZE * alloc_size;
	douonP->douon = (struct douon *)calloc(
				douonP->count, sizeof(struct douon) );
	bzero( douonP->douon, douonP->count * sizeof(struct douon) );
	if ( old )
	    free( old );
    }
    douonP->use_count = count;
    return;
}

Private void
sj3_SetBunsetu( bunsetuP, bunsetu, count )
    sj3_bunsetu		*bunsetuP;
    struct bunsetu	*bunsetu;
    int			 count;
{
    struct bunsetu	*old = bunsetuP->bun;
    int			 needed_size = count;

#define SJ3_BUNSETU_ALLOC_SIZE	8
    if ( count == 0 )
	return;
    if ( needed_size > bunsetuP->count ) {
	int	alloc_size = needed_size / SJ3_BUNSETU_ALLOC_SIZE;
	alloc_size += ((needed_size % SJ3_BUNSETU_ALLOC_SIZE) ? 1 : 0);
	bunsetuP->count = SJ3_BUNSETU_ALLOC_SIZE * alloc_size;
	bunsetuP->bun = (struct bunsetu *)
			    calloc( bunsetuP->count, sizeof(struct bunsetu) );
	if ( old )
	    free( old );
    }
    bunsetuP->use_count = count;
    bcopy( bunsetu, bunsetuP->bun, count * sizeof(struct bunsetu) );
    return;
}

Private void
sj3_SetFeedback( s, feedbackP, bunsetuP )
    iml_session_t	*s;
    sj3_feedback	*feedbackP;
    sj3_bunsetu		*bunsetuP;
{
    int			 i;
    XIMFeedback		 feedback;
    struct bunsetu	*bunsetu = bunsetuP->bun;

    feedbackP->use_count = 0;
    for ( i = 0; i < s->bnum; i++, bunsetu++ ) {
	if ( i == s->current_bnum ) {
	    feedback = XIMReverse;
	} else {
	    feedback = XIMUnderline;
	}
	sj3_CopyFeedbacks( feedbackP, feedback, bunsetu->destlen / 2 );
    }
    return;
}

Private void
ReSetBunsetu( convP, bun, bnum )
    sj3_SJISstring	*convP;
    struct bunsetu	*bun;
    int			 bnum;
{
    int		i;
    int		loc = 0;

    for ( i = 0; i < bnum; i++, bun++ ) {
	bun->deststr = (unsigned char *)(convP->string + loc );
	loc += bun->destlen;
    }
    return;
}

Private void
SetPreedit( s, bunsetuP, sjis, size )
    iml_session_t	*s;
    sj3_bunsetu		*bunsetuP;
    String		 sjis;
    int			 size;
{
    int			 i;
    int			 wordCount;
    int			 allocSize = strlen( s->convString.string ) + size + 1;
    int			 eucAllocSize = allocSize * sizeof(wchar_t);
    String		 convBuff, convPtr;
    XIMFeedback		 feedback;
    wchar_t		*eucBuff;
    struct bunsetu	*bunsetu = bunsetuP->bun;

    convPtr = convBuff = (String )calloc( allocSize, sizeof(char) );
    bzero( convBuff, allocSize );

    s->preeditString.use_count = 0;
    s->preeditFeedback.use_count = 0;
    for ( i = 0; i < s->bnum; i++, bunsetu++ ) {
	if ( i == s->current_bnum ) {
	    bcopy( sjis, convPtr, size );
	    convPtr += size;
	    bunsetu->destlen = size;
	    sj3_CopyFeedbacks( &s->preeditFeedback, XIMReverse, size / 2 );
	} else {
	    bcopy( bunsetu->deststr, convPtr, bunsetu->destlen );
	    convPtr += bunsetu->destlen;
	    sj3_CopyFeedbacks( &s->preeditFeedback,
				XIMUnderline, bunsetu->destlen / 2 );
	}
    }
    *convPtr = '\0';
    s->convString.use_byte = 0;
    sj3_AddSJISStrings( &s->convString, convBuff, strlen( convBuff ) );
    ReSetBunsetu( &s->convString, bunsetuP->bun, s->bnum );
    eucBuff = (wchar_t *)calloc( eucAllocSize, sizeof(wchar_t) );
    bzero( eucBuff, eucAllocSize );
    wordCount = ShiftJisToEuc( convBuff,
			strlen(convBuff), eucBuff, eucAllocSize );
    sj3_AddPreeditStrings( &s->preeditString, eucBuff, wordCount );

    return;
}

Private void
GetDouon( douP, curBun )
    sj3_douon		*douP;
    struct bunsetu	*curBun;
{
    int		allocSize = curBun->srclen + 1;
    int		word_count;
    String	sjisBuff;
    String	p;

    sjisBuff = (String )calloc( allocSize, sizeof(char) );
    bcopy( curBun->srcstr, sjisBuff, curBun->srclen );
    p = sjisBuff;
    p += curBun->srclen;
    *p = '\0';

    if ( ( word_count = sj3_douoncnt( sjisBuff ) ) < 0 ) {
	    return;
    }
    sj3_AllocDouonBuff( douP, word_count );
    sj3_getdouon( sjisBuff, douP->douon );
    free( sjisBuff );
    return;
}

Private int
ShiftJisToEuc( sjis, sjlen, euc, euclen )
    unsigned char	*sjis;
    int			 sjlen;
    wchar_t		*euc;
    int			 euclen;
{
    int		 	 i, j;
    wchar_t	 	 tmp;
    unsigned short	*mbsbuf, *mbsbufP;

    mbsbufP = mbsbuf = (unsigned short *)calloc( sjlen + 1, sizeof(unsigned short) );
    bzero( mbsbuf, sjlen + 1 );
    i = j = 0;
    while ( euclen > i && sjlen > j ) {
	if ( ( sjis[j] >= 0x81 && sjis[j] <= 0x9f ) ||
	     ( sjis[j] >= 0xe0 && sjis[j] <= 0xfc ) ) {
	    tmp = sjis_to_jis( ((((unsigned short)( sjis[j] )) << 8 ) & 0xff00 )
		    | ((unsigned short)( sjis[j+1] ) & 0x00ff ) );
	    tmp |= 0x8080;
	    *mbsbuf = tmp;
	    mbsbuf++; i++; j += sizeof(unsigned short);
	} else {
	    *mbsbuf = ((unsigned short)( sjis[j] )) & 0x00ff;
	    mbsbuf++; i++; j++;
	}
    }
    mbstowcs( euc, (char *)mbsbufP, sjlen );
    free( mbsbufP );
    return i;
}

Private wchar_t
sjis_to_jis( code )
    unsigned short	code;
{
    unsigned char	 high, low;

    high = ( code >> 8 ) & 0xff;
    low  = code & 0xff;

    if ( high < 0x81 || ( high > 0x9f && high < 0xe0 ) || high > 0xfc ||
	 low < 0x40 || low == 0x7f || low > 0xfc ) {
	return -1;
    }
    high -= ( high >= 0xe0 ? 0xb0 : 0x70 );
    high <<= 1;
    if ( low < 0x80 )
	low++;
    if ( low >= 0x9f ) {
	high |= 0x01;
	low  -= 0x7e;
    } else {
	low -= 0x20;
    }
    --high;
    return ( ( ( high << 8 ) & 0xff00 ) | low );
}

Private int
EucToShiftJis( euc, euc_num, sjis, sjis_size )
    wchar_t		*euc;
    int			 euc_num;
    unsigned char	*sjis;
    int			 sjis_size;
{
    int			 i, j = 0;
    int			 size = euc_num * sizeof(unsigned short) + 1;
    unsigned short	 tmp;
    unsigned short	*mbsbuf, *mbsbufP;

    mbsbufP = mbsbuf = (unsigned short *)calloc( size, sizeof(unsigned short) );
    wcstombs( (char *)mbsbufP, euc, euc_num * sizeof(wchar_t) );
    for ( i = 0; i < euc_num; i++, mbsbuf++ ) {
	if ( *mbsbuf < 0xff ) {
	    if ( j < sjis_size ) {
		sjis[j] = (unsigned char)(*mbsbuf);
		j++;
	    } else {
		free( (char *)mbsbufP );
		return j;
	    }
	} else {
	    if ( j + 1 < sjis_size ) {
		tmp = *mbsbuf & 0x7f7f;
		tmp = jis_to_sjis( tmp );
		sjis[j] = (unsigned char)( tmp >> 8 );
		sjis[j+1] = (unsigned char)( tmp & 0xff );
		j += 2;
	    } else {
		free( (char *)mbsbufP );
		return j;
	    }
	}
    }
    free( (char *)mbsbufP );
    return j;
}

Private unsigned short
jis_to_sjis( code )
    unsigned short	code;
{
    unsigned short	a, b;

    a = ( code >> 8 ) & 0xff;
    b = code & 0xff;
    if ( a < 0x21 || a > 0x7e && a < 0x21 || b > 0x7e ) {
	return -1;
    }
    a++;
    b += ( a & 0x01 ? 0x7e : 0x20 );
    if ( b < 0x80 ) --b;
    a = a >> 1;
    a += 0x70;
    if ( a >= 0x9f ) a += 0x40;

    return ( ( ( a << 8 ) & 0xff00 ) | b );
}

Private String
HiraganaToKatakana( str, size )
    String	str;
    int		size;
{
    static unsigned short	euc;
    String			p, pP;

    pP = p = (String )calloc( size + 1, sizeof(char) );
    for ( ; *str; ) {
	euc = ( *str++ & 0xff ) << 8;
	euc += ( *str++ & 0xff );
	if ( euc >= 0xa4a1 && euc <= 0xa4f4 )
	    euc += 0x100;
	*p = ( euc >> 8 ) & 0xff; p++;
	*p = euc & 0xff; p++;
    }
    *p = '\0';
    return pP;
}
